#ifndef _URCU_ARCH_UATOMIC_HPPA_H
#define _URCU_ARCH_UATOMIC_HPPA_H

#include <urcu/compiler.h>
#include <urcu/system.h>

#define UATOMIC_HAS_ATOMIC_SHORT
#include <urcu/uatomic/generic.h>

#endif /* _URCU_ARCH_UATOMIC_HPPA_H */
